UPDATED_CONTRACT_KEY = "updated_contract"
OLD_CONTRACT_KEY = "old_contract"

CONTRACT_SENDER_ID = "senderid"
CONTRACT_RECEIVER_ID = "receiverid"
CONTRACT_PRODUCT_NAME = "productname"
CONTRACT_FIELD_DELIVERY_INTERVAL_IN_STEPS = "deliveryintervalinsteps"